///////////////////////////////////////////////////////////////////////////////
// Created on: April 2000
// Created by: Ronald Hordijk (hk@dgmr.nl)
//
// Storage class for the generate polygons.
// When Polygons are added, normals are calculated and a check is done if points
//   of the polygon are already use.
// The normal on a point is calculated bij middeling the normals on the polygons
//   using the point.
//
// The class is filled bij CalcThead using the fullowing sequence
//   Startfill
//     several CellChange
//       per cellchange several AddPolygon
//   Endfill
// Knowledge of the way the grid is traversed and which cell is being calculated
// is used for optimising the test if points can be reused. (witch took a very long
//        time testing all previous points)
//
// Paintthread uses the function converting the data to Opengl-format showing the data.
//
// Because the data is used by two threads the class is made 'thread-save' using as
// critalsection.
//
// Because of the large number of possible polygons > 500K this class might
// demand a large chunk of memory.
//////////////////////////////////////////////////////////////////////////////
unit FaceList;

interface

uses
   syncobjs, vector;

type
  TFace = record
  p1,p2,p3: integer;
  Normal: TVector3;
  end;

type
  TFaceList = class
  private
    LastPoint: array of array of Integer;
    LastFace: array[0..1] of Integer;
    CurrentLayer: Integer;
    CurrentRow: Integer;

    Points: array of TVector3;
    Normals: array of TVector3;
    nrPoints: Integer;
    FaceList: array of TFace;
    nrFaceList: integer;
    LastDrawPolygonCount : Integer;
    bBusyFilling: boolean;

    CS : TCriticalSection;

    function AddPoint(Point: TVector3): Integer;
    function AddFace(p1,p2,p3: Integer; normal: TVector3): Integer;
  public
    constructor Create;
    Destructor  Destroy; override;

    procedure StartFill(const Layers, Rows, Colomns: Integer);
    procedure CellChange(const Layer, Row, Colomn: Integer);
    function  AddPolygon(p1,p2,p3,dir: TVector3): Integer;
    procedure EndFill;

    function  PartPolygonsToGl(Count: Integer; Normal: boolean): Boolean;
    procedure StartPartToGl;
    procedure FastToGl(Count: Integer);
    procedure PointsToGl;

    function  FaceCount: Integer;
    function  PointCount: Integer;
    function  BusyFilling: Boolean;
  end;

implementation

uses
  opengl12, math;

const
  LISTSTEPSIZE = 16384;
  Epsilon = 1e-6;

///////////////////////////////////////////////////////////////////////////////
constructor TFaceList.Create;
begin
  CS := TCriticalSection.Create;
  bBusyFilling := False;
  nrPoints := 0;
  nrFaceList := 0;
end;

///////////////////////////////////////////////////////////////////////////////
destructor TFaceList.Destroy;
begin
  CS.Free;

  inherited;
end;

///////////////////////////////////////////////////////////////////////////////
function TFaceList.AddFace(p1, p2, p3: Integer; normal: TVector3): Integer;
begin
  if (nrFaceList mod LISTSTEPSIZE) = 0 then begin
    SetLength(FaceList, nrFaceList + LISTSTEPSIZE);
  end;

  FaceList[nrFaceList].p1 := p1;
  FaceList[nrFaceList].p2 := p2;
  FaceList[nrFaceList].p3 := p3;
  FaceList[nrFaceList].normal := normal;

  Result := nrFaceList;
  Inc(nrFaceList);
end;

///////////////////////////////////////////////////////////////////////////////
function TFaceList.AddPoint(Point: TVector3): Integer;
var
  i: integer;
begin
  // Try and find is point was already used
  for i := LastPoint[(CurrentLayer-1) mod 2][Max(CurrentRow - 1, 0)] to
           LastPoint[(CurrentLayer-1) mod 2][Min(CurrentRow + 1, High(LastPoint[0]))] - 1 do begin
    if (abs(Points[i][0] - Point[0]) < Epsilon) and
        (abs(Points[i][1] - Point[1]) < Epsilon) and
         (abs(Points[i][2] - Point[2]) < Epsilon) then begin
      Result := i;
      Exit;
    end;
  end;

  for i := LastPoint[(CurrentLayer) mod 2][Max(CurrentRow - 1, 0)] to
           nrPoints - 1 do begin
    if (abs(Points[i][0] - Point[0]) < Epsilon) and
        (abs(Points[i][1] - Point[1]) < Epsilon) and
         (abs(Points[i][2] - Point[2]) < Epsilon) then begin
      Result := i;
      Exit;
    end;
  end;

  if (nrPoints mod LISTSTEPSIZE) = 0 then begin
    SetLength(Points, nrPoints + LISTSTEPSIZE);
    SetLength(Normals, nrPoints + LISTSTEPSIZE);
    for i:= nrPoints to nrPoints + LISTSTEPSIZE - 1 do
      Normals[i] := Vector3(0,0,0);
  end;

  Points[nrPoints] := Point;

  Result := nrPoints;
  Inc(nrPoints);
end;

///////////////////////////////////////////////////////////////////////////////
procedure TFaceList.StartFill(const Layers, Rows, Colomns: Integer);
var
 i: integer;
begin
  CS.Acquire;

  Setlength(LastPoint,2,Rows);
  for i:= 0 to Rows - 1 do begin
    LastPoint[0, i] := 0;
    LastPoint[1, i] := 0;
  end;

  LastFace[0] := 0;
  LastFace[1] := 0;

  CurrentLayer := 0;
  CurrentRow := 0;
  nrPoints := 0;
  nrFaceList := 0;

  bBusyFilling := True;

  CS.Release;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TFaceList.CellChange(const Layer, Row, Colomn: Integer);
begin
  CS.Acquire;
  CurrentLayer := Layer;
  CurrentRow := Row;

  LastPoint[CurrentLayer mod 2][CurrentRow] := nrPoints;
  LastFace[CurrentLayer mod 2] := nrFaceList;
  CS.Release;
end;

///////////////////////////////////////////////////////////////////////////////
function TFaceList.AddPolygon(p1, p2, p3, dir: TVector3): Integer;
var
  v1,v2,n: TVector3;
begin
  CS.Acquire;

  v1 := Vector3Sub(p1,p2);
  v2 := Vector3Sub(p1,p3);

  n[0] := v1[1] * v2[2] - v1[2] * v2[1];
  n[1] := v1[2] * v2[0] - v1[0] * v2[2];
  n[2] := v1[0] * v2[1] - v1[1] * v2[0];

  n := Vector3Normalize(n);

  if (Vector3Angle(n,dir) > 1.57) then
    Result := AddFace(AddPoint(p1), AddPoint(p2), AddPoint(p3), Vector3Mult(n, -1))
  else
    Result := AddFace(AddPoint(p2), AddPoint(p1), AddPoint(p3), n);

  Normals[FaceList[Result].p1] := Vector3add(Normals[FaceList[Result].p1], FaceList[Result].Normal);
  Normals[FaceList[Result].p2] := Vector3add(Normals[FaceList[Result].p2], FaceList[Result].Normal);
  Normals[FaceList[Result].p3] := Vector3add(Normals[FaceList[Result].p3], FaceList[Result].Normal);

  CS.Release;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TFaceList.EndFill;
var
  i: integer;
begin
  CS.Acquire;
  for i := 0 to nrPoints - 1 do
    Normals[i] := Vector3Normalize(Normals[i]);

  bBusyFilling := False;
  CS.Release;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TFaceList.StartPartToGl;
begin
  LastDrawPolygonCount := 0;
end;

///////////////////////////////////////////////////////////////////////////////
function TFaceList.PartPolygonsToGl(Count: Integer; Normal: boolean): Boolean;
var
  i: integer;
  NewLastDrawPolygonCount: integer;
begin
  CS.Acquire;

  glBegin(GL_TRIANGLES);
  if (nrFaceList - 1) <= (LastDrawPolygonCount + Count) then begin
    NewLastDrawPolygonCount := nrFaceList - 1;
    Result := False;
  end else begin
    NewLastDrawPolygonCount := LastDrawPolygonCount + Count;
    Result := True;
  end;

  for i := LastDrawPolygonCount to NewLastDrawPolygonCount do begin
    if not Normal then begin
      glVertex3d(Points[FaceList[i].p1][0], Points[FaceList[i].p1][1], Points[FaceList[i].p1][2]);
      glVertex3d(Points[FaceList[i].p2][0], Points[FaceList[i].p2][1], Points[FaceList[i].p2][2]);
      glVertex3d(Points[FaceList[i].p3][0], Points[FaceList[i].p3][1], Points[FaceList[i].p3][2]);
    end else begin
      glTexCoord3d(Points[FaceList[i].p1][0] / 2 + 0.5,
                   Points[FaceList[i].p1][1] / 2 + 0.5,
                   Points[FaceList[i].p1][2] / 2 + 0.5);
      glNormal3d(Normals[FaceList[i].p1][0], Normals[FaceList[i].p1][1], Normals[FaceList[i].p1][2]);
      glVertex3d(Points[FaceList[i].p1][0], Points[FaceList[i].p1][1], Points[FaceList[i].p1][2]);
      glTexCoord3d(Points[FaceList[i].p2][0] / 2 + 0.5,
                   Points[FaceList[i].p2][1] / 2 + 0.5,
                   Points[FaceList[i].p2][2] / 2 + 0.5);
      glNormal3d(Normals[FaceList[i].p2][0], Normals[FaceList[i].p2][1], Normals[FaceList[i].p2][2]);
      glVertex3d(Points[FaceList[i].p2][0], Points[FaceList[i].p2][1], Points[FaceList[i].p2][2]);
      glTexCoord3d(Points[FaceList[i].p3][0] / 2 + 0.5,
                   Points[FaceList[i].p3][1] / 2 + 0.5,
                   Points[FaceList[i].p3][2] / 2 + 0.5);
      glNormal3d(Normals[FaceList[i].p3][0], Normals[FaceList[i].p3][1], Normals[FaceList[i].p3][2]);
      glVertex3d(Points[FaceList[i].p3][0], Points[FaceList[i].p3][1], Points[FaceList[i].p3][2]);
    end;
  end;
  glEnd;

  if Result then begin
    LastDrawPolygonCount := NewLastDrawPolygonCount
  end else begin
    if bBusyFilling then
      LastDrawPolygonCount := 0
    else
      LastDrawPolygonCount := nrFaceList - 1;
  end;

  CS.Release;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TFaceList.FastToGl(Count: Integer);
var
  i,j: integer;
begin
  CS.Acquire;

  if nrPoints < Count then begin
    PointsToGl;
    CS.Release;
    Exit;
  end;

  glBegin(GL_POINTS);
  for j := 0 to Count-1 do begin
    i := (int64(j) * (nrPoints - 1)) div Count;

    glVertex3d(Points[i][0], Points[i][1], Points[i][2]);
  end;
  glEnd();

  CS.Release
end;

///////////////////////////////////////////////////////////////////////////////
procedure TFaceList.PointsToGl;
var
  i: integer;
begin
  CS.Acquire;

  glBegin(GL_POINTS);
  for i := 0 to nrPoints - 1 do begin
    glVertex3d(Points[i][0], Points[i][1], Points[i][2]);
  end;
  glEnd();

  CS.Release;
end;

///////////////////////////////////////////////////////////////////////////////
function TFaceList.FaceCount: Integer;
begin
  Result := nrFaceList;
end;

///////////////////////////////////////////////////////////////////////////////
function TFaceList.PointCount: Integer;
begin
  Result := nrPoints;
end;

///////////////////////////////////////////////////////////////////////////////
function TFaceList.BusyFilling: Boolean;
begin
  Result := bBusyFilling;
end;

///////////////////////////////////////////////////////////////////////////////
end.
