///////////////////////////////////////////////////////////////////////////////
// Created on: April 2000
// Created by: Ronald Hordijk (hk@dgmr.nl)
//
// Calculate the Julia Value (Number of interations) at a particular point in
// space for several julia functions.
// This code is base on the julia code from povray.
///////////////////////////////////////////////////////////////////////////////

unit JuliaCalc;

interface

uses
  Vector;

type
  TCalcParam = record
    Param:    array[0..3] of Double;
    NrInter:  Integer;
    CalcType: Integer;
    Border:   Boolean;
     // Flag to set how the border value should be calculated:
     // True  - Only within or without the border value.
     // False - Also how far from the border.
  end;

type
  TJuliaCalcFunc = function (v: TVector3): Integer of Object;

type
  TJuliaCalculator = class
  private
    CalcParam: TCalcParam; // Parameters neccesary for the calculation

  public
    function SetParameters(const ACalcParam: TCalcParam): TJuliaCalcFunc;
    procedure Complex_Mult(var x,y: double; ax,ay,bx,by: double);
    procedure Complex_Sqrt(var x,y: double; ax,ay: double);
    procedure Complex_Log(var x,y: double; ax,ay: double);

    function  CalcJuliaQSqr(v: TVector3): Integer;
    function  CalcJuliaQCube(v: TVector3): Integer;
    function  CalcJuliaHSqr(v: TVector3): Integer;
    function  CalcJuliaHCube(v: TVector3): Integer;
    function  CalcJuliaHReciprocal(v: TVector3): Integer;
    function  CalcJuliaHExp(v: TVector3): Integer;
    function  CalcJuliaHLog(v: TVector3): Integer;
    function  CalcJuliaHSin(v: TVector3): Integer;
    function  CalcJuliaHASin(v: TVector3): Integer;
    function  CalcJuliaHCos(v: TVector3): Integer;
    function  CalcJuliaHACos(v: TVector3): Integer;
    function  CalcJuliaHTan(v: TVector3): Integer;
    function  CalcJuliaHATan(v: TVector3): Integer;
    function  CalcJuliaHSinH(v: TVector3): Integer;
    function  CalcJuliaHASinH(v: TVector3): Integer;
    function  CalcJuliaHCosH(v: TVector3): Integer;
    function  CalcJuliaHACosH(v: TVector3): Integer;
    function  CalcJuliaHTanH(v: TVector3): Integer;
    function  CalcJuliaHATanH(v: TVector3): Integer;
  end;

implementation

uses
  Math;

///////////////////////////////////////////////////////////////////////////////
function TJuliaCalculator.SetParameters(const ACalcParam: TCalcParam): TJuliaCalcFunc;
begin
  CalcParam := ACalcParam;

  case CalcParam.CalcType of
   0: Result := CalcJuliaQSQr;
   1: Result := CalcJuliaQCube;
   2: Result := CalcJuliaHSqr;
   3: Result := CalcJuliaHCube;
   4: Result := CalcJuliaHExp;
   5: Result := CalcJuliaHSin;
   6: Result := CalcJuliaHCos;
   7: Result := CalcJuliaHSinH;
   8: Result := CalcJuliaHASinH;
   9: Result := CalcJuliaHCosH;
  10: Result := CalcJuliaHACosH;
  11: Result := CalcJuliaHReciprocal;
  12: Result := CalcJuliaHLog;
  13: Result := CalcJuliaHASin;
  14: Result := CalcJuliaHACos;
  15: Result := CalcJuliaHATan;
  16: Result := CalcJuliaHATanH;
  17: Result := CalcJuliaHTan;
  18: Result := CalcJuliaHTanH;
  end;
end;

///////////////////////////////////////////////////////////////////////////////
function TJuliaCalculator.CalcJuliaQSqr(v: TVector3): Integer;
var
  x,y,z,t,d,x2,w: double;
  i: integer;
  EndLoop: integer;
begin
  Result := 2 * CalcParam.NrInter;

  x := v[0];
  y := v[1];
  z := v[2];
  w := 0;

  if CalcParam.Border then
    EndLoop := CalcParam.NrInter
  else
    EndLoop := 2 * CalcParam.NrInter;

  for i := 1 to EndLoop do begin
    d := y * y + z * z + w * w;

    x2 := x * x;

    if (d + x2 > 4) then begin
      Result := i;
      Exit;
    end;

    t := 2.0 * x;

    x := x2 - d  + CalcParam.Param[0];
    y := y * t   + CalcParam.Param[1];
    z := z * t   + CalcParam.Param[2];
    w := w * t   + CalcParam.Param[3];
  end;
end;

///////////////////////////////////////////////////////////////////////////////
function TJuliaCalculator.CalcJuliaQCube(v: TVector3): Integer;
var
  x,y,z,t,d,x2,w: double;
  i: integer;
  EndLoop: integer;
begin
  Result := 2 * CalcParam.NrInter;

  x := v[0];
  y := v[1];
  z := v[2];
  w := 0;

  if CalcParam.Border then
    EndLoop := CalcParam.NrInter
  else
    EndLoop := 2 * CalcParam.NrInter;

  for i := 1 to EndLoop do begin
    d := y * y + z * z + w * w;

    x2 := x * x;

    if (d + x2 > 4) then begin
      Result := i;
      Exit;
    end;

    t := 3.0 * x2 - d;

    x := x * (x2 - 3.0 * d)  + CalcParam.Param[0];
    y := y * t   + CalcParam.Param[1];
    z := z * t   + CalcParam.Param[2];
    w := w * t   + CalcParam.Param[3];
  end;
end;

///////////////////////////////////////////////////////////////////////////////
function TJuliaCalculator.CalcJuliaHSqr(v: TVector3): Integer;
var
  x,y,z,w,yz,xw,t1,t2,t3: double;
  i: integer;
  EndLoop: integer;
begin
  Result := 2 * CalcParam.NrInter;

  x := v[0];
  y := v[1];
  z := v[2];
  w := 0;

  if CalcParam.Border then
    EndLoop := CalcParam.NrInter
  else
    EndLoop := 2 * CalcParam.NrInter;

  for i := 1 to EndLoop do begin
    yz := y * y + z * z;
    xw := x * x + w * w;

    if ((yz + xw) > 16) then begin
      Result := i;
      Exit;
    end;

    t1 := 2 * (x * y - z * w);
    t2 := 2 * (x * z - w * y);
    t3 := 2 * (x * w + y * z);

    x := xw - yz  + CalcParam.Param[0];
    y := t1 + CalcParam.Param[1];
    z := t2 + CalcParam.Param[2];
    w := t3 + CalcParam.Param[3];
  end;
end;

///////////////////////////////////////////////////////////////////////////////
function TJuliaCalculator.CalcJuliaHCube(v: TVector3): Integer;
var
  x,y,z,w,d,t1,t2,t3,t4: double;
  i: integer;
  EndLoop: integer;
begin
  Result := 2 * CalcParam.NrInter;

  x := v[0];
  y := v[1];
  z := v[2];
  w := 0;

  if CalcParam.Border then
    EndLoop := CalcParam.NrInter
  else
    EndLoop := 2 * CalcParam.NrInter;

  for i := 1 to EndLoop do begin
    d := x * x + y * y + z * z + w * w;

    if (d > 16) then begin
      Result := i;
      Exit;
    end;

    Complex_Mult(t1, t2, (x - w), (y + z), (x - w), (y + z));
    Complex_Mult(t1, t2, t1, t2, (x - w), (y + z));
    Complex_Mult(t3, t3, (x + w), (y - z), (x + w), (y - z));
    Complex_Mult(t3, t4, t3, t4, (x + w), (y - z));

    x := 0.5 * (t1 + t3) + CalcParam.Param[0];
    y := 0.5 * (t2 + t4) + CalcParam.Param[1];
    z := 0.5 * (t2 - t4) + CalcParam.Param[2];
    w := 0.5 * (t3 - t1) + CalcParam.Param[3];
  end;
end;

///////////////////////////////////////////////////////////////////////////////
function TJuliaCalculator.CalcJuliaHACos(v: TVector3): Integer;
var
  x,y,z,w,d,t1,t2,t3,t4: double;
  i: integer;
  EndLoop: integer;
begin
  Result := 2 * CalcParam.NrInter;

  x := v[0];
  y := v[1];
  z := v[2];
  w := 0;

  if CalcParam.Border then
    EndLoop := CalcParam.NrInter
  else
    EndLoop := 2 * CalcParam.NrInter;

  for i := 1 to EndLoop do begin
    d := x * x + y * y + z * z + w * w;

    if (d > 16) then begin
      Result := i;
      Exit;
    end;

    Complex_Mult(t2, t1, x - w, y + z, x - w, y + z);
    t2 := t2 - 1;
    Complex_Sqrt(t2, t1, t2, t1);
    t2 := t2 + x - w;
    t1 := t1 + y + z;
    Complex_Log(t2, t1, t2, t1);
    t2 := -t2;

    Complex_Mult(t4, t3, x + w, y - z, x + w, y - z);
    t4 := t4 - 1;
    Complex_Sqrt(t4, t3, t4, t3);
    t4 := t4 + x + w;
    t3 := t3 + y - z;
    Complex_Log(t4, t3, t4, t3);
    t4 := -t4;

    x := 0.5 * (t1 + t3) + CalcParam.Param[0];
    y := 0.5 * (t2 + t4) + CalcParam.Param[1];
    z := 0.5 * (t2 - t4) + CalcParam.Param[2];
    w := 0.5 * (t3 - t1) + CalcParam.Param[3];
  end;
end;

///////////////////////////////////////////////////////////////////////////////
function TJuliaCalculator.CalcJuliaHACosH(v: TVector3): Integer;
var
  x,y,z,w,d,t1,t2,t3,t4: double;
  i: integer;
  EndLoop: integer;
begin
  Result := 2 * CalcParam.NrInter;

  x := v[0];
  y := v[1];
  z := v[2];
  w := 0;

  if CalcParam.Border then
    EndLoop := CalcParam.NrInter
  else
    EndLoop := 2 * CalcParam.NrInter;

  for i := 1 to EndLoop do begin
    d := x * x + y * y + z * z + w * w;

    if (d > 16) then begin
      Result := i;
      Exit;
    end;
    Complex_Mult(t1, t2, x - w, y + z, x - w, y + z);
    t1 := t1 - 1;
    Complex_Sqrt(t1, t2, t1, t2);
    t1 := t1 + x - w;
    t2 := t2 + y + z;
    Complex_Log(t1, t2, t1, t2);

    Complex_Mult(t3, t4, x + w, y - z, x + w, y - z);
    t1 := t1 - 1;
    Complex_Sqrt(t3, t4, t3, t4);
    t3 := t3 + x + w;
    t4 := t4 + y - z;
    Complex_Log(t3, t4, t3, t4);

    x := 0.5 * (t1 + t3) + CalcParam.Param[0];
    y := 0.5 * (t2 + t4) + CalcParam.Param[1];
    z := 0.5 * (t2 - t4) + CalcParam.Param[2];
    w := 0.5 * (t3 - t1) + CalcParam.Param[3];
  end;
end;

///////////////////////////////////////////////////////////////////////////////
function TJuliaCalculator.CalcJuliaHASin(v: TVector3): Integer;
var
  x,y,z,w,d,t1,t2,t3,t4: double;
  i: integer;
  EndLoop: integer;
begin
  Result := 2 * CalcParam.NrInter;

  x := v[0];
  y := v[1];
  z := v[2];
  w := 0;

  if CalcParam.Border then
    EndLoop := CalcParam.NrInter
  else
    EndLoop := 2 * CalcParam.NrInter;

  for i := 1 to EndLoop do begin
    d := x * x + y * y + z * z + w * w;

    if (d > 16) then begin
      Result := i;
      Exit;
    end;

    Complex_Mult(t2, t1, x - w, y + z, x - w, y + z);
    t2 := t2 - 1;
    t1 := -t1;
    Complex_Sqrt(t2, t1, t2, t1);
    t2 := t2 + x - w;
    t1 := t1 + y + z;
    Complex_Log(t2, t1, t2, t1);
    t2 := -t2;

    Complex_Mult(t4, t3, x + w, y - z, x + w, y - z);
    t4 := t4 - 1;
    t3 := -t3;
    Complex_Sqrt(t4, t3, t4, t3);
    t4 := t4 + x + w;
    t3 := t3 + y - z;
    Complex_Log(t4, t3, t4, t3);
    t4 := -t4;

    x := 0.5 * (t1 + t3) + CalcParam.Param[0];
    y := 0.5 * (t2 + t4) + CalcParam.Param[1];
    z := 0.5 * (t2 - t4) + CalcParam.Param[2];
    w := 0.5 * (t3 - t1) + CalcParam.Param[3];
  end;
end;

///////////////////////////////////////////////////////////////////////////////
function TJuliaCalculator.CalcJuliaHASinH(v: TVector3): Integer;
var
  x,y,z,w,d,t1,t2,t3,t4: double;
  i: integer;
  EndLoop: integer;
begin
  Result := 2 * CalcParam.NrInter;

  x := v[0];
  y := v[1];
  z := v[2];
  w := 0;

  if CalcParam.Border then
    EndLoop := CalcParam.NrInter
  else
    EndLoop := 2 * CalcParam.NrInter;

  for i := 1 to EndLoop do begin
    d := x * x + y * y + z * z + w * w;

    if (d > 16) then begin
      Result := i;
      Exit;
    end;
    Complex_Mult(t1, t2, x - w, y + z, x - w, y + z);
    t1 := t1 + 1;
    Complex_Sqrt(t1, t2, t1, t2);
    t1 := t1 + x - w;
    t2 := t2 + y + z;
    Complex_Log(t1, t2, t1, t2);

    Complex_Mult(t3, t4, x + w, y - z, x + w, y - z);
    t1 := t1 + 1;
    Complex_Sqrt(t3, t4, t3, t4);
    t3 := t3 + x + w;
    t4 := t4 + y - z;
    Complex_Log(t3, t4, t3, t4);

    x := 0.5 * (t1 + t3) + CalcParam.Param[0];
    y := 0.5 * (t2 + t4) + CalcParam.Param[1];
    z := 0.5 * (t2 - t4) + CalcParam.Param[2];
    w := 0.5 * (t3 - t1) + CalcParam.Param[3];
  end;
end;

///////////////////////////////////////////////////////////////////////////////
function TJuliaCalculator.CalcJuliaHATan(v: TVector3): Integer;
//var
//  x,y,z,w,d,t1,t2,t3,t4: double;
//  i: integer;
//  EndLoop: integer;
begin
  Result := 2 * CalcParam.NrInter;
{
  x := v[0];
  y := v[1];
  z := v[2];
  w := 0;

  if CalcParam.Border then
    EndLoop := CalcParam.NrInter
  else
    EndLoop := 2 * CalcParam.NrInter;

  for i := 1 to EndLoop do begin
    d := x * x + y * y + z * z + w * w;

    if (d > 16) then begin
      Result := i;
      Exit;
    end;

    x := 0.5 * (t1 + t3) + CalcParam.Param[0];
    y := 0.5 * (t2 + t4) + CalcParam.Param[1];
    z := 0.5 * (t2 - t4) + CalcParam.Param[2];
    w := 0.5 * (t3 - t1) + CalcParam.Param[3];
  end;
}
end;

///////////////////////////////////////////////////////////////////////////////
function TJuliaCalculator.CalcJuliaHATanH(v: TVector3): Integer;
//var
//  x,y,z,w,d,t1,t2,t3,t4: double;
//  i: integer;
//  EndLoop: integer;
begin
  Result := 2 * CalcParam.NrInter;
{
  x := v[0];
  y := v[1];
  z := v[2];
  w := 0;

  if CalcParam.Border then
    EndLoop := CalcParam.NrInter
  else
    EndLoop := 2 * CalcParam.NrInter;

  for i := 1 to EndLoop do begin
    d := x * x + y * y + z * z + w * w;

    if (d > 16) then begin
      Result := i;
      Exit;
    end;

    x := 0.5 * (t1 + t3) + CalcParam.Param[0];
    y := 0.5 * (t2 + t4) + CalcParam.Param[1];
    z := 0.5 * (t2 - t4) + CalcParam.Param[2];
    w := 0.5 * (t3 - t1) + CalcParam.Param[3];
  end;
}
end;

///////////////////////////////////////////////////////////////////////////////
function TJuliaCalculator.CalcJuliaHCos(v: TVector3): Integer;
var
  x,y,z,w,d,t1,t2,t3,t4: double;
  i: integer;
  EndLoop: integer;
begin
  Result := 2 * CalcParam.NrInter;

  x := v[0];
  y := v[1];
  z := v[2];
  w := 0;

  if CalcParam.Border then
    EndLoop := CalcParam.NrInter
  else
    EndLoop := 2 * CalcParam.NrInter;

  for i := 1 to EndLoop do begin
    d := x * x + y * y + z * z + w * w;

    if (d > 16) then begin
      Result := i;
      Exit;
    end;

    t1 := cos(x - w) * cosh(y + z);
    t2 := -sin(x - w) * sinh(y + z);
    t3 := cos(x + w) * cosh(y - z);
    t4 := -sin(x + w) * sinh(y - z);

    x := 0.5 * (t1 + t3) + CalcParam.Param[0];
    y := 0.5 * (t2 + t4) + CalcParam.Param[1];
    z := 0.5 * (t2 - t4) + CalcParam.Param[2];
    w := 0.5 * (t3 - t1) + CalcParam.Param[3];
  end;
end;

///////////////////////////////////////////////////////////////////////////////
function TJuliaCalculator.CalcJuliaHCosH(v: TVector3): Integer;
var
  x,y,z,w,d,t1,t2,t3,t4: double;
  i: integer;
  EndLoop: integer;
begin
  Result := 2 * CalcParam.NrInter;

  x := v[0];
  y := v[1];
  z := v[2];
  w := 0;

  if CalcParam.Border then
    EndLoop := CalcParam.NrInter
  else
    EndLoop := 2 * CalcParam.NrInter;

  for i := 1 to EndLoop do begin
    d := x * x + y * y + z * z + w * w;

    if (d > 16) then begin
      Result := i;
      Exit;
    end;

    t1 := cosh(x - w) * cos(y + z);
    t2 := sinh(x - w) * sin(y + z);
    t3 := cosh(x + w) * cos(y - z);
    t4 := sinh(x + w) * sin(y - z);

    x := 0.5 * (t1 + t3) + CalcParam.Param[0];
    y := 0.5 * (t2 + t4) + CalcParam.Param[1];
    z := 0.5 * (t2 - t4) + CalcParam.Param[2];
    w := 0.5 * (t3 - t1) + CalcParam.Param[3];
  end;
end;

///////////////////////////////////////////////////////////////////////////////
function TJuliaCalculator.CalcJuliaHExp(v: TVector3): Integer;
var
  x,y,z,w,d,t1,t2,t3,t4: double;
  i: integer;
  EndLoop: integer;
begin
  Result := 2 * CalcParam.NrInter;

  x := v[0];
  y := v[1];
  z := v[2];
  w := 0;

  if CalcParam.Border then
    EndLoop := CalcParam.NrInter
  else
    EndLoop := 2 * CalcParam.NrInter;

  for i := 1 to EndLoop do begin
    d := x * x + y * y + z * z + w * w;

    if (d > 16) then begin
      Result := i;
      Exit;
    end;

    t1 := exp(x - w) * cos(y + z);
    t2 := exp(x - w) * sin(y + z);
    t3 := exp(x + w) * cos(y - z);
    t4 := exp(x + w) * sin(y - z);

    x := 0.5 * (t1 + t3) + CalcParam.Param[0];
    y := 0.5 * (t2 + t4) + CalcParam.Param[1];
    z := 0.5 * (t2 - t4) + CalcParam.Param[2];
    w := 0.5 * (t3 - t1) + CalcParam.Param[3];
  end;
end;

///////////////////////////////////////////////////////////////////////////////
function TJuliaCalculator.CalcJuliaHLog(v: TVector3): Integer;
var
  x,y,z,w,d,t1,t2,t3,t4: double;
  i: integer;
  EndLoop: integer;
begin
  Result := 2 * CalcParam.NrInter;

  x := v[0];
  y := v[1];
  z := v[2];
  w := 0;

  if CalcParam.Border then
    EndLoop := CalcParam.NrInter
  else
    EndLoop := 2 * CalcParam.NrInter;

  for i := 1 to EndLoop do begin
    d := x * x + y * y + z * z + w * w;

    if (d > 16) then begin
      Result := i;
      Exit;
    end;

    t1 := log10(sqrt((x - w) * (x - w) + (y + z) * (y + z)));
    t2 := arctan2((y + z), (x - w));
    t3 := log10(sqrt((x + w) * (x + w) + (y - z) * (y - z)));
    t4 := arctan2((y - z), (x + w));

    x := 0.5 * (t1 + t3) + CalcParam.Param[0];
    y := 0.5 * (t2 + t4) + CalcParam.Param[1];
    z := 0.5 * (t2 - t4) + CalcParam.Param[2];
    w := 0.5 * (t3 - t1) + CalcParam.Param[3];
  end;
end;

///////////////////////////////////////////////////////////////////////////////
function TJuliaCalculator.CalcJuliaHReciprocal(v: TVector3): Integer;
var
  x,y,z,w,d, d2, d3 ,t1,t2,t3,t4: double;
  i: integer;
  EndLoop: integer;
begin
  Result := 2 * CalcParam.NrInter;

  x := v[0];
  y := v[1];
  z := v[2];
  w := 0;

  if CalcParam.Border then
    EndLoop := CalcParam.NrInter
  else
    EndLoop := 2 * CalcParam.NrInter;

  for i := 1 to EndLoop do begin
    d := x * x + y * y + z * z + w * w;

    if (d > 16) then begin
      Result := i;
      Exit;
    end;

    d2 := ((x - w) * (x - w) + (y + z) * (y + z)) * ((x + w) * (x + w) + (y - z) * (y - z));

    if (d2 = 0.0) then
      Exit;

    d3 := x * w - y * z;

    t1 := (x * d - 2 * w * d3) / d2;
    t2 := (-y * d - 2 * z * d3) / d2;
    t3 := (-z * d - 2 * y * d3) / d2;
    t4 := (w * d - 2 * x * d3) / d2;

    x := t1 + CalcParam.Param[0];
    y := t2 + CalcParam.Param[1];
    z := t3 + CalcParam.Param[2];
    w := t4 + CalcParam.Param[3];
  end;
end;

///////////////////////////////////////////////////////////////////////////////
function TJuliaCalculator.CalcJuliaHSin(v: TVector3): Integer;
var
  x,y,z,w,d,t1,t2,t3,t4: double;
  i: integer;
  EndLoop: integer;
begin
  Result := 2 * CalcParam.NrInter;

  x := v[0];
  y := v[1];
  z := v[2];
  w := 0;

  if CalcParam.Border then
    EndLoop := CalcParam.NrInter
  else
    EndLoop := 2 * CalcParam.NrInter;

  for i := 1 to EndLoop do begin
    d := x * x + y * y + z * z + w * w;

    if (d > 16) then begin
      Result := i;
      Exit;
    end;

    t1 := sin(x - w) * cosh(y + z);
    t2 := cos(x - w) * sinh(y + z);
    t3 := sin(x + w) * cosh(y - z);
    t4 := cos(x + w) * sinh(y - z);

    x := 0.5 * (t1 + t3) + CalcParam.Param[0];
    y := 0.5 * (t2 + t4) + CalcParam.Param[1];
    z := 0.5 * (t2 - t4) + CalcParam.Param[2];
    w := 0.5 * (t3 - t1) + CalcParam.Param[3];
  end;
end;

///////////////////////////////////////////////////////////////////////////////
function TJuliaCalculator.CalcJuliaHSinH(v: TVector3): Integer;
var
  x,y,z,w,d,t1,t2,t3,t4: double;
  i: integer;
  EndLoop: integer;
begin
  Result := 2 * CalcParam.NrInter;

  x := v[0];
  y := v[1];
  z := v[2];
  w := 0;

  if CalcParam.Border then
    EndLoop := CalcParam.NrInter
  else
    EndLoop := 2 * CalcParam.NrInter;

  for i := 1 to EndLoop do begin
    d := x * x + y * y + z * z + w * w;

    if (d > 16) then begin
      Result := i;
      Exit;
    end;

    t1 := sinh(x - w) * cos(y + z);
    t2 := cosh(x - w) * sin(y + z);
    t3 := sinh(x + w) * cos(y - z);
    t4 := cosh(x + w) * sin(y - z);

    x := 0.5 * (t1 + t3) + CalcParam.Param[0];
    y := 0.5 * (t2 + t4) + CalcParam.Param[1];
    z := 0.5 * (t2 - t4) + CalcParam.Param[2];
    w := 0.5 * (t3 - t1) + CalcParam.Param[3];
  end;
end;

///////////////////////////////////////////////////////////////////////////////
function TJuliaCalculator.CalcJuliaHTan(v: TVector3): Integer;
var
  x,y,z,w,d,t1,t2,t3,t4: double;
  i: integer;
  EndLoop: integer;
begin
  Result := 2 * CalcParam.NrInter;

  x := v[0];
  y := v[1];
  z := v[2];
  w := 0;

  if CalcParam.Border then
    EndLoop := CalcParam.NrInter
  else
    EndLoop := 2 * CalcParam.NrInter;

  for i := 1 to EndLoop do begin
    d := x * x + y * y + z * z + w * w;

    if (d > 16) then begin
      Result := i;
      Exit;
    end;

    d  := cos(2 *(x - w)) + cosh(2 *(y + z));
    if d = 0 then Exit;
    t1 := sin(2 * (x - w))/ d;
    t2 := sinh(2 * (y + z))/ d;
    d  := cos(2 *(x + w)) + cosh(2 *(y - z));
    if d = 0 then Exit;
    t3 := sin(2 * (x + w))/ d;
    t4 := sinh(2 * (y - z))/ d;

    x := 0.5 * (t1 + t3) + CalcParam.Param[0];
    y := 0.5 * (t2 + t4) + CalcParam.Param[1];
    z := 0.5 * (t2 - t4) + CalcParam.Param[2];
    w := 0.5 * (t3 - t1) + CalcParam.Param[3];
  end;
end;

///////////////////////////////////////////////////////////////////////////////
function TJuliaCalculator.CalcJuliaHTanH(v: TVector3): Integer;
var
  x,y,z,w,d,t1,t2,t3,t4: double;
  i: integer;
  EndLoop: integer;
begin
  Result := 2 * CalcParam.NrInter;

  x := v[0];
  y := v[1];
  z := v[2];
  w := 0;

  if CalcParam.Border then
    EndLoop := CalcParam.NrInter
  else
    EndLoop := 2 * CalcParam.NrInter;

  for i := 1 to EndLoop do begin
    d := x * x + y * y + z * z + w * w;

    if (d > 16) then begin
      Result := i;
      Exit;
    end;

    d  := cosh(2 *(x - w)) + cos(2 *(y + z));
    if d = 0 then Exit;
    t1 := sinh(2 * (x - w))/ d;
    t2 := sin(2 * (y + z))/ d;
    d  := cosh(2 *(x + w)) + cos(2 *(y - z));
    if d = 0 then Exit;
    t3 := sinh(2 * (x + w))/ d;
    t4 := sin(2 * (y - z))/ d;

    x := 0.5 * (t1 + t3) + CalcParam.Param[0];
    y := 0.5 * (t2 + t4) + CalcParam.Param[1];
    z := 0.5 * (t2 - t4) + CalcParam.Param[2];
    w := 0.5 * (t3 - t1) + CalcParam.Param[3];
  end;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TJuliaCalculator.Complex_Mult(var x, y: double; ax, ay, bx,
  by: double);
begin
  x := ax * bx - ay * by;
  y := ax * by + ay * bx;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TJuliaCalculator.Complex_Sqrt(var x, y: double; ax, ay: double);
var
  mag, theta: double;
begin
  if(ax = 0.0) and (ay = 0.0) then begin
    x := 0; y := 0;
  end else begin
    mag   := sqrt(sqrt(ax * ax + ay * ay));
    theta := arctan2(ay, ax) / 2;
    y := mag * sin(theta);
    x := mag * cos(theta);
  end;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TJuliaCalculator.Complex_Log(var x, y: double; ax, ay: double);
begin
  if(ax = 0.0) and (ay = 0.0) then begin
    x := 0; y := 0;
  end else begin
    x := log10(sqrt(ax * ax + ay * ay));
    y := arctan2(ay,ax);
  end;
end;

///////////////////////////////////////////////////////////////////////////////
end.
