///////////////////////////////////////////////////////////////////////////////
// Created on: April 2000
// Created by: Ronald Hordijk (hk@dgmr.nl)
//
// Draws the data from facelist an the zoom-box.
// Because of possible large number of polygons > 500K the painting might take
// several seconds. To prevent the delay the painting is divided into parts.
// After each part the screen is updated. Also the painting is moved to a seperate
// thread so the the main form remains responsive.
//
// Some simple opengl code is used for the painting.
///////////////////////////////////////////////////////////////////////////////
unit PaintThread;

interface

uses
   Classes, Graphics, Windows, FaceList, opengl12, Coordinates, Vector;

type
  TPaintParam = record
    RC:            HGLRC;
    DC:            HDC;

    ObjectColor:   TColor;
    BackColor:     TColor;
    Fog:           Boolean;
    DrawMode:      Integer;
    QuickDrawSize: Integer;
    Azimuth,
    AngleHeight:   Double;
    QuickDraw:     Boolean;
    DrawBox:       Boolean;

    ViewUp:        TVector3;
    ViewRight:     TVector3;
    ViewLook:      TVector3;
  end;

type
  TpaintThread = class(TThread)
  private
    FaceList:     TFaceList;
    Coordinates:  TCoordinates;
    PaintParam:   TPaintParam;
    procedure DrawBorder;
    procedure DrawBox(v1, v2: TVector3);
    procedure DrawSquare(v1, v2: TVector3; DrawMode: integer);
  protected
    procedure Execute; override;
  public
    constructor Create(AFaceList: TFaceList; APaintParam: TPaintParam; ACoordinates: TCoordinates);
  end;

implementation

{ TpaintThread }

///////////////////////////////////////////////////////////////////////////////
constructor TpaintThread.Create(AFaceList: TFaceList; APaintParam: TPaintParam;
                                                   ACoordinates: TCoordinates);
begin
  FaceList := AFaceList;
  Coordinates := ACoordinates;
  PaintParam := APaintParam;

  inherited Create(False);
end;

///////////////////////////////////////////////////////////////////////////////
procedure TpaintThread.DrawBox(v1,v2: TVector3);
begin
  glbegin(GL_LINE_STRIP);
  glVerTex3d(v1[0], v1[1], v1[2]);
  glVerTex3d(v1[0], v1[1], v2[2]);
  glVerTex3d(v1[0], v2[1], v2[2]);
  glVerTex3d(v1[0], v2[1], v1[2]);
  glVerTex3d(v1[0], v1[1], v1[2]);
  glVerTex3d(v2[0], v1[1], v1[2]);
  glVerTex3d(v2[0], v2[1], v1[2]);
  glVerTex3d(v2[0], v2[1], v2[2]);
  glVerTex3d(v2[0], v1[1], v2[2]);
  glVerTex3d(v2[0], v1[1], v1[2]);
  glEnd;

  glbegin(GL_LINES);
  glVerTex3d(v1[0], v1[1], v2[2]);
  glVerTex3d(v2[0], v1[1], v2[2]);
  glVerTex3d(v1[0], v2[1], v2[2]);
  glVerTex3d(v2[0], v2[1], v2[2]);
  glVerTex3d(v1[0], v2[1], v1[2]);
  glVerTex3d(v2[0], v2[1], v1[2]);
  glEnd;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TpaintThread.DrawSquare(v1,v2: TVector3; DrawMode: integer);
const
  STEPSIZE = 10;
var
  i: integer;
begin
 glbegin(GL_LINES);
  Case DrawMode of
  0:
    for i := 0 to STEPSIZE do begin
      glVertex3d(v1[0], v1[1], v1[2] + i * (v2[2] - v1[2]) /STEPSIZE);
      glVertex3d(v1[0], v2[1], v1[2] + i * (v2[2] - v1[2]) /STEPSIZE);
      glVertex3d(v1[0], v1[1] + i * (v2[1] - v1[1]) /STEPSIZE, v1[2]);
      glVertex3d(v1[0], v1[1] + i * (v2[1] - v1[1]) /STEPSIZE, v2[2]);
    end;
  1:
    for i := 0 to STEPSIZE do begin
      glVertex3d(v1[0], v1[1], v1[2] + i * (v2[2] - v1[2]) /STEPSIZE);
      glVertex3d(v2[0], v1[1], v1[2] + i * (v2[2] - v1[2]) /STEPSIZE);
      glVertex3d(v1[0] + i * (v2[0] - v1[0]) /STEPSIZE, v1[1], v1[2]);
      glVertex3d(v1[0] + i * (v2[0] - v1[0]) /STEPSIZE, v1[1], v2[2]);
    end;
  2:
    for i := 0 to STEPSIZE do begin
      glVertex3d(v1[0], v1[1] + i * (v2[1] - v1[1]) /STEPSIZE, v1[2]);
      glVertex3d(v2[0], v1[1] + i * (v2[1] - v1[1]) /STEPSIZE, v1[2]);
      glVertex3d(v1[0] + i * (v2[0] - v1[0]) /STEPSIZE, v1[1], v1[2]);
      glVertex3d(v1[0] + i * (v2[0] - v1[0]) /STEPSIZE, v2[1], v1[2]);
    end;
  else ;
  end;
  glEnd;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TpaintThread.DrawBorder;
const
  Epsilon = 0.00001;
  STEPSIZE = 10;
var
  LocViewStart, LocViewStop,
  LocNewViewStart, LocnewViewStop: TVector3;
begin
  glDisable(GL_FOG);
  glEnable (GL_BLEND);
  glEnable (GL_LINE_SMOOTH);

  LocViewStart    := Coordinates.JuliaToViewCoord(Coordinates.Start);
  LocViewStop     := Coordinates.JuliaToViewCoord(Coordinates.Stop);
  LocNewViewStart := Coordinates.JuliaToViewCoord(Coordinates.NewStart);
  LocnewViewStop  := Coordinates.JuliaToViewCoord(Coordinates.NewStop);

  glColor3d(1,1,1);
  DrawBox(LocViewStart, LocViewStop);

  DrawBox(LocNewViewStart, LocnewViewStop);

  glColor4d(1, 1, 1, 0.4);

  if Abs(Coordinates.Start[0] - Coordinates.NewStart[0]) > Epsilon then begin
    DrawSquare(Vector3(LocNewViewStart[0], LocViewStart[1], LocViewStart[2]),
               Vector3(LocNewViewStart[0], LocViewStop[1],  LocViewStop[2]),0);
  end;

  if Abs(Coordinates.Stop[0] - Coordinates.NewStop[0]) > Epsilon then begin
    DrawSquare(Vector3(LocNewViewStop[0], LocViewStart[1], LocViewStart[2]),
               Vector3(LocNewViewStop[0], LocViewStop[1],  LocViewStop[2]),0);
  end;

  if Abs(Coordinates.Start[1] - Coordinates.NewStart[1]) > Epsilon then begin
    DrawSquare(Vector3(LocViewStart[0], LocNewViewStart[1], LocViewStart[2]),
               Vector3(LocViewStop[0],  LocNewViewStart[1],  LocViewStop[2]),1);
  end;

  if Abs(Coordinates.Stop[1] - Coordinates.NewStop[1]) > Epsilon then begin
    DrawSquare(Vector3(LocViewStart[0], LocNewViewStop[1], LocViewStart[2]),
               Vector3(LocViewStop[0],  LocNewViewStop[1],  LocViewStop[2]),1);
  end;

  if Abs(Coordinates.Start[2] - Coordinates.NewStart[2]) > Epsilon then begin
    DrawSquare(Vector3(LocViewStart[0], LocViewStart[1], LocNewViewStart[2]),
               Vector3(LocViewStop[0],  LocViewStop[1],  LocNewViewStart[2]), 2);
  end;

  if Abs(Coordinates.Stop[2] - Coordinates.NewStop[2]) > Epsilon then begin
    DrawSquare(Vector3(LocViewStart[0], LocViewStart[1], LocNewViewStop[2]),
               Vector3(LocViewStop[0],  LocViewStop[1],  LocNewViewStop[2]), 2);
  end;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TpaintThread.Execute;
Var
  DrawMode : integer;
begin
  ActivateRenderingContext(PaintParam.DC, PaintParam.RC);

  glClearColor(GetRValue(PaintParam.BackColor)/255,
               GetGValue(PaintParam.BackColor)/255,
               GetBValue(PaintParam.BackColor)/255, 0);

  glClear(GL_DEPTH_BUFFER_BIT or GL_COLOR_BUFFER_BIT);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity;
  gluLookAt( PaintParam.Viewlook[0] * 6, PaintParam.Viewlook[1] * 6, PaintParam.Viewlook[2] * 6,
             0, 0, 0,
             PaintParam.ViewUp[0], PaintParam.ViewUp[1], PaintParam.ViewUp[2]);
//  gluLookAt( 6, 0, 0,  0, 0, 0,  0, 0, 1);
//  glRotatef(PaintParam.AngleHeight,0,1,0);
//  glRotatef(PaintParam.Azimuth,0,0,1);

  glcolor3ub(GetRValue(PaintParam.ObjectColor),
             GetGValue(PaintParam.ObjectColor),
             GetBValue(PaintParam.ObjectColor));

  if PaintParam.fog or PaintParam.QuickDraw or FaceList.BusyFilling then
    glEnable(GL_FOG)
  else
    glDisable(GL_FOG);

  glDisable (GL_BLEND);
  glDisable (GL_LINE_SMOOTH);

  if PaintParam.QuickDraw then begin
    FaceList.FastToGl(PaintParam.QuickDrawSize);
  end else begin
    FaceList.StartPartToGl;

    if FaceList.BusyFilling then
      DrawMode := 2
    else
      DrawMode := PaintParam.DrawMode;

    case DrawMode of
    0: // points
       FaceList.PointsToGl;
    1: begin  // lines
         glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
         While (FaceList.PartPolygonsToGl(10000, False)) and
                (Not Terminated) do begin
           SwapBuffers(PaintParam.DC);
           glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
         end;
         glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
       end;
    2: begin  // polygons
         While (FaceList.PartPolygonsToGl(10000, False) or (FaceList.BusyFilling)) and
                (Not Terminated) do begin
           SwapBuffers(PaintParam.DC);
         end;
       end;
    3: begin
         glEnable(GL_LIGHT1);
         glEnable(GL_LIGHTING);
         glEnable(GL_COLOR_MATERIAL);
         While (FaceList.PartPolygonsToGl(10000, True)) and
                (Not Terminated) do begin
           SwapBuffers(PaintParam.DC);
         end;
         glDisable(GL_COLOR_MATERIAL);
         glDisable(GL_LIGHTING);
         glDisable(GL_LIGHT1);
       end;
    4: begin
         glEnable(GL_LIGHT1);
         glEnable(GL_LIGHTING);
         glEnable(GL_COLOR_MATERIAL);
         GLEnable(GL_TEXTURE_2D);
         GLEnable(GL_BLEND);
//         GLBlendFunc(GL_ONE,GL_ONE);

//         glEnable(GL_TEXTURE_GEN_S);
//         glEnable(GL_TEXTURE_GEN_T);

//         glEnable(GL_TEXTURE_GEN_Q);

         glcolor3ub(255,255,255);


         While (FaceList.PartPolygonsToGl(10000, True)) and
                (Not Terminated) do begin
           SwapBuffers(PaintParam.DC);
         end;
//         glDisable(GL_TEXTURE_GEN_S);
//         glDisable(GL_TEXTURE_GEN_T);

//         glDisable(GL_TEXTURE_GEN_Q);


         GLDisable(GL_TEXTURE_2D);
         GLDisable(GL_BLEND);
         glDisable(GL_COLOR_MATERIAL);
         glDisable(GL_LIGHTING);
         glDisable(GL_LIGHT1);
       end;
    end;
  end;

  // zoom box
  if PaintParam.DrawBox or PaintParam.QuickDraw then
    DrawBorder;

  SwapBuffers(PaintParam.DC);
  DeactivateRenderingContext;
end;

///////////////////////////////////////////////////////////////////////////////
end.
