///////////////////////////////////////////////////////////////////////////////
// Created on: April 2000
// Created by: Ronald Hordijk (hk@dgmr.nl)
//
// A Trackbar for two values.
///////////////////////////////////////////////////////////////////////////////
unit RangeSel;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls;

type
  TRangeSelector = class(TGraphicControl)
  private
    FBorderStyle: TBorderStyle;
    FBackColor: TColor;
    FForeColor: TColor;
    FLowValue:  Integer;
    FHighValue: Integer;
    FDragging:   Boolean;
    FOnValueChanged: TNotifyEvent;

    procedure SetBorderStyle(const Value: TBorderStyle);
    procedure SetForeColor(const Value: TColor);
    procedure SetBackColor(const Value: TColor);
    procedure SetLowValue(const Value: Integer);
    procedure SetHighValue(const Value: Integer);
    procedure SetNewValue(x: integer);
  protected
    { Protected declarations }
    procedure Paint; override;

    procedure MouseDown(Sender: TObject; Button: TMouseButton;
               Shift: TShiftState; X, Y: Integer); reintroduce;
    procedure MouseUp(Sender: TObject; Button: TMouseButton;
               Shift: TShiftState; X, Y: Integer); reintroduce;
    procedure MouseMove(Sender: TObject;
               Shift: TShiftState; X, Y: Integer); reintroduce;
  public
    { Public declarations }

    constructor Create(AOwner: TComponent); override;

    property Dragging :Boolean read FDragging;
  published
    { Published declarations }

    property BackColor: TColor read FBackColor write SetBackColor default clWhite;
    property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle default bsSingle;
    property Color;
    property Enabled;
    property ForeColor: TColor read FForeColor write SetForeColor default clBlue;
    property Visible;
    property LowValue: Integer read FLowValue write SetLowValue default 0;
    property HighValue: Integer read FHighValue write SetHighValue default 1000;

    property OnValueChanged: TNotifyEvent read FOnValueChanged write FOnValueChanged;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Samples', [TRangeSelector]);
end;

{ TRangeSelector }

///////////////////////////////////////////////////////////////////////////////
constructor TRangeSelector.Create(AOwner: TComponent);
begin
  inherited;

  Width  := 100;
  Height := 20;

  FBorderStyle := bsSingle;
  FForeColor   := clBlue;
  FBackColor   := clWhite;

  FLowValue := 0;
  FHighValue := 1000;

  FDragging := False;
  OnMouseMove := MouseMove;
  OnMouseDown := MouseDown;
  OnMouseUp   := MouseUp;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TRangeSelector.MouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if Shift <> [ssLeft] then
    Exit;

  FDragging := True
end;

///////////////////////////////////////////////////////////////////////////////
procedure TRangeSelector.MouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  if Shift <> [ssLeft] then
    Exit;

  SetNewValue(X);
end;

///////////////////////////////////////////////////////////////////////////////
procedure TRangeSelector.MouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if not FDragging then
    Exit;

  FDragging := False;
  SetNewValue(X);

  if Assigned(FOnValueChanged) then
    FOnValueChanged(Self);
end;

///////////////////////////////////////////////////////////////////////////////
procedure TRangeSelector.Paint;
var
  PaintRect: TRect;
  TheImage: TBitmap;
begin
  TheImage := TBitmap.Create;
  try
    TheImage.Height := Height;
    TheImage.Width := Width;

    TheImage.Canvas.Pen.Color := clBlack;
    TheImage.Canvas.Brush.Color := FBackColor;
    PaintRect := ClientRect;
    if FBorderStyle = bsSingle then
      TheImage.Canvas.Rectangle(PaintRect)
    else
      TheImage.Canvas.FillRect(PaintRect);

    inflateRect(PaintRect,0,-1);

    if FBorderStyle = bsSingle then begin
      PaintRect.Left := Round((LowValue * (Width - 2)) / 1000) + 1;
      PaintRect.Right := Round((HighValue * (Width - 2)) / 1000) + 1;
    end else begin
      PaintRect.Left := Round((LowValue * Width) / 1000);
      PaintRect.Right := Round((HighValue * Width) / 1000);
    end;

    TheImage.Canvas.Brush.Color := FForeColor;
    TheImage.Canvas.FillRect(PaintRect);

    Canvas.CopyMode := cmSrcCopy;
    Canvas.Draw(0, 0, TheImage);
  finally
    TheImage.Free;
  end;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TRangeSelector.SetBackColor(const Value: TColor);
begin
  if Value <> FBackColor then begin
    FBackColor := Value;
    refresh;
  end;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TRangeSelector.SetBorderStyle(const Value: TBorderStyle);
begin
  if Value <> FBorderStyle then begin
    FBorderStyle := Value;
    refresh;
  end;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TRangeSelector.SetForeColor(const Value: TColor);
begin
  if Value <> FForeColor then begin
    FForeColor := Value;
    refresh;
  end;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TRangeSelector.SetHighValue(const Value: Integer);
begin
  if Value <> HighValue then begin
    if Value < LowValue then begin
      HighValue := LowValue;
      LowValue := Value;
    end else if Value = LowValue then begin
      if Value = 1000 then
        LowValue := Value - 1
      else
        HighValue := Value + 1;
    end else begin
      FHighValue := Value;
      refresh;
      if Assigned(FOnValueChanged) then
        FOnValueChanged(Self);
    end;
  end;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TRangeSelector.SetLowValue(const Value: Integer);
begin
  if Value <> LowValue then begin
    if Value > HighValue then begin
      LowValue := HighValue;
      HighValue := Value;
    end else if Value = HighValue then begin
      if Value = 0 then
        HighValue := Value + 1
      else
        LowValue := Value - 1;
    end else begin
      FLowValue := Value;
      refresh;
      if Assigned(FOnValueChanged) then
        FOnValueChanged(Self);
    end;
  end;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TRangeSelector.SetNewValue(X: integer);
var
  pos: integer;
begin
  if FBorderStyle = bsSingle then begin
    if X < 1 then begin
      pos := 0;
    end else if X > (width - 1) then begin
      pos := 1000;
    end else begin
      pos := Round(((x - 1) * 1000) / (width - 2));
    end;
  end else begin
    if X < 0 then begin
      pos := 0;
    end else if X > width then begin
      pos := 1000;
    end else begin
      pos := Round((x * 1000) / width);
    end;
  end;

  if Pos > HighValue then begin
    HighValue := Pos
  end else if Pos < LowValue then begin
    LowValue := Pos
  end else if Abs(Pos - LowValue) < Abs(Pos - HighValue) then begin
    LowValue := Pos
  end else begin
    HighValue := Pos;
  end;
end;

end.
